/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2000 Ian Zimmerman <itz@speakeasy.org> */

static const char rcsid[] = "$Id: cook.c,v 1.12 2002/06/16 04:36:05 itz Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ggi/gii.h>
#include <sys/time.h>
#include <limits.h>
#include <assert.h>

#include "gim_event.h"
#include "cook.h"
#include "linuxonly.h"

struct timeval double_time = {0, 400000};

int threshold = 0;
int scale_up = 1;
int scale_down = 1;

int acx;
int acy;
unsigned cook_state = 0;

int button_sequence[9] = { 1, 2, 3, 4, 5, 6, 7, 8, 9 };

static int
transform (int x, int threshold, int* ac)
{
  int xabs;
  int xsign;

  xabs = abs(x);
  xsign = (x == 0 ? 0 : x/xabs);

  if (xabs > threshold) {
    xabs -= threshold;
    xabs *= scale_up;
    xabs /= scale_down;
  } else {
    xabs = 0;
  } /*if*/

  if (xabs > 0) return xabs * xsign;
  if (ac != 0) {
    *ac += x;
    return transform(*ac, threshold, 0);
  } else {
    return 0;
  } /*if*/
}

static int
timeval_subtract (result, x, y)
     struct timeval *result, *x, *y;
{
  /* Perform the carry for the later subtraction by updating Y. */
  if (x->tv_usec < y->tv_usec) {
    int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
    y->tv_usec -= 1000000 * nsec;
    y->tv_sec += nsec;
  }
  if (x->tv_usec - y->tv_usec > 1000000) {
    int nsec = (x->tv_usec - y->tv_usec) / 1000000;
    y->tv_usec += 1000000 * nsec;
    y->tv_sec -= nsec;
  }
     
  /* Compute the time remaining to wait.
     `tv_usec' is certainly positive. */
  if (result != 0) {
    result->tv_sec = x->tv_sec - y->tv_sec;
    result->tv_usec = x->tv_usec - y->tv_usec;
  } /*if*/
  /* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}

int
cook_event (const gii_event* gii_ev, struct gim_event_t* gim_ev)
{
  struct timeval my_double_time;
  struct timeval result_time;
  struct timeval my_last_down_time;
                                /* absolute position tracking */
  static int x, y, z;
                                /* these are for finding multi-clicks */
  static int clicks = 0;
  static int last_down_button = -1;
  static struct timeval last_down_time;
  static int dragged = 0;
  
  int scaled_gii_x, scaled_gii_y;
  int my_button;
  int wx, wy;

  gim_ev->vt = current_vt();
  get_window_size(&wx, &wy, gim_ev->vt);
  gim_ev->time = gii_ev->any.time;
  gim_ev->modifiers = get_kbd_modifiers();

  switch (gii_ev->any.type) {
  case evPtrAbsolute:
    /* any absolute move clears small move accumulators */
    acx = acy = 0;
    /* scale down gii coordinates. no threshold for absolute coordinates */
    scaled_gii_x = transform(gii_ev->pmove.x, 0, 0);
    scaled_gii_y = transform(gii_ev->pmove.y, 0, 0);
    if (scaled_gii_x == x && scaled_gii_y == y && gii_ev->pmove.z == z) return 0;

    gim_ev->state = cook_state;
    if (cook_state == 0) {
      gim_ev->kind = gim_move_ev;
      clicks = 0;
    } else {
      gim_ev->kind = gim_drag_ev;
      dragged = 1;
      if (clicks > 0) gim_ev->kind |= (gim_single_ev << (clicks - 1));
    } /*if*/
    gim_ev->u.d.margin = gim_none_mg;
                                /* fit into window */
    if (scaled_gii_x < 0) {
      gim_ev->x = 0;
      gim_ev->u.d.margin |= gim_left_mg;
    } else if (scaled_gii_x >= wx) {
      gim_ev->x = wx - 1;
      gim_ev->u.d.margin |= gim_right_mg;
    } else {
      gim_ev->x = scaled_gii_x;
    } /*if*/
    if (scaled_gii_y < 0) {
      gim_ev->y = 0;
      gim_ev->u.d.margin |= gim_top_mg;
    } else if (scaled_gii_y >= wy) {
      gim_ev->y = wy - 1;
      gim_ev->u.d.margin |= gim_bottom_mg;
    } else {
      gim_ev->y = scaled_gii_y;
    } /*if*/
    gim_ev->z = gii_ev->pmove.z;

    gim_ev->u.d.x = gim_ev->x - x;
    gim_ev->u.d.y = gim_ev->y - y;
    gim_ev->u.d.z = gim_ev->z - z;
    x = gim_ev->x;
    y = gim_ev->y;
    z = gim_ev->z;
    break;
  case evPtrRelative:
    /* scale down gii coordinates */
    scaled_gii_x = transform(gii_ev->pmove.x, threshold, &acx);
    scaled_gii_y = transform(gii_ev->pmove.y, threshold, &acy);
    if (scaled_gii_x == 0 && scaled_gii_y == 0 && gii_ev->pmove.z == 0) {
      return 0;
    } else {
      acx = acy = 0;
    } /*if*/
    gim_ev->state = cook_state;
    if (cook_state == 0) {
      gim_ev->kind = gim_move_ev;
      clicks = 0;
    } else {
      gim_ev->kind = gim_drag_ev;
      dragged = 1;
      if (clicks > 0) gim_ev->kind |= (gim_single_ev << (clicks - 1));
    } /*if*/
    gim_ev->u.d.margin = gim_none_mg;
                                /* fit into window */
    if (x + scaled_gii_x < 0) {
      gim_ev->x = 0;
      gim_ev->u.d.margin |= gim_left_mg;
    } else if (x + scaled_gii_x >= wx) {
      gim_ev->x = wx - 1;
      gim_ev->u.d.margin |= gim_right_mg;
    } else {
      gim_ev->x = x + scaled_gii_x;
    } /*if*/
    if (y + scaled_gii_y < 0) {
      gim_ev->y = 0;
      gim_ev->u.d.margin |= gim_top_mg;
    } else if (y + scaled_gii_y >= wy) {
      gim_ev->y = wy - 1;
      gim_ev->u.d.margin |= gim_bottom_mg;
    } else {
      gim_ev->y = y + scaled_gii_y;
    } /*if*/
    gim_ev->z = gii_ev->pmove.z;

    gim_ev->u.d.x = gim_ev->x - x;
    gim_ev->u.d.y = gim_ev->y - y;
    gim_ev->u.d.z = gim_ev->z - z;
    x = gim_ev->x;
    y = gim_ev->y;
    z = gim_ev->z;
    break;
  case evPtrButtonRelease:
    /* any button event clears small move accumulators */
    acx = acy = 0;
    /* translate button number */
    my_button = (gii_ev->pbutton.button > 9 ? gii_ev->pbutton.button :
                 button_sequence[gii_ev->pbutton.button - 1]);

    gim_ev->kind = gim_up_ev;
    if (dragged) gim_ev->kind |= gim_dragged_click_ev;
    if (my_button != last_down_button) {
      clicks = 0;
    } /*if*/
    if (clicks > 0) gim_ev->kind |= (gim_single_ev << (clicks - 1));
    if (my_button <= CHAR_BIT * sizeof(cook_state)) {
      cook_state &= ~(1 << (my_button - 1));
      if (cook_state == 0) dragged = 0;
    } /*if*/
    gim_ev->state = cook_state;
    gim_ev->u.b.which = my_button;
    gim_ev->u.b.repeat = clicks;
    gim_ev->x = x;
    gim_ev->y = y;
    gim_ev->z = z;
    break;
  case evPtrButtonPress:
    /* any button event clears small move accumulators */
    acx = acy = 0;
    /* translate button number */
    my_button = (gii_ev->pbutton.button > 9 ? gii_ev->pbutton.button :
                 button_sequence[gii_ev->pbutton.button - 1]);

    gim_ev->kind = gim_down_ev;
    if (my_button <= CHAR_BIT * sizeof(cook_state)) {
      cook_state |= (1 << (my_button - 1));
    } /*if*/
    gim_ev->state = cook_state;
    gim_ev->u.b.which = my_button;
    gim_ev->x = x;
    gim_ev->y = y;
    gim_ev->z = z;

    my_double_time = double_time;
    my_last_down_time = last_down_time;
    timeval_subtract(&result_time, &gim_ev->time, &my_double_time);
    if (my_button != last_down_button ||
        !timeval_subtract(0, &result_time, &my_last_down_time)) {
      clicks = 1;
    } else {
      clicks += 1;
      if (clicks > 3) {
        clicks = 1;
      } /*if*/
    } /*if*/
    gim_ev->kind |= (gim_single_ev << (clicks - 1));

    gim_ev->u.b.repeat = clicks;
    last_down_button = gim_ev->u.b.which;
    last_down_time = gim_ev->time;
    break;

  default:
    assert(0);
  } /*switch*/

  return 1;
}
